import datajoint as dj
import numpy as np
import pathlib
from datetime import datetime
import uuid
import hashlib
import importlib
import inspect

from . import scan, find_full_path
from .scan import get_imaging_root_data_dir

schema = dj.schema()

_linking_module = None


def activate(imaging_schema_name, scan_schema_name=None, *,
             create_schema=True, create_tables=True, linking_module=None):
    """
    activate(imaging_schema_name, *, scan_schema_name=None, create_schema=True, create_tables=True, linking_module=None)
        :param imaging_schema_name: schema name on the database server to activate the `imaging` module
        :param scan_schema_name: schema name on the database server to activate the `scan` module
         - may be omitted if the `scan` module is already activated
        :param create_schema: when True (default), create schema in the database if it does not yet exist.
        :param create_tables: when True (default), create tables in the database if they do not yet exist.
        :param linking_module: a module name or a module containing the
         required dependencies to activate the `imaging` module:
         + all that are required by the `scan` module
    """

    if isinstance(linking_module, str):
        linking_module = importlib.import_module(linking_module)
    assert inspect.ismodule(linking_module),\
        "The argument 'dependency' must be a module's name or a module"

    global _linking_module
    _linking_module = linking_module

    scan.activate(scan_schema_name, create_schema=create_schema,	
                  create_tables=create_tables, linking_module=linking_module)
    schema.activate(imaging_schema_name, create_schema=create_schema,
                    create_tables=create_tables, add_objects=_linking_module.__dict__)

# -------------- Table declarations --------------


@schema
class ProcessingMethod(dj.Lookup):
    definition = """  #  Method, package, analysis suite used for processing of calcium imaging data (e.g. Suite2p, CaImAn, etc.)
    processing_method: char(8)
    ---
    processing_method_desc: varchar(1000)
    """

    contents = [('suite2p', 'suite2p analysis suite'),
                ('caiman', 'caiman analysis suite')]


@schema
class ProcessingParamSet(dj.Lookup):
    definition = """  #  Parameter set used for processing of calcium imaging data
    paramset_idx:  smallint
    ---
    -> ProcessingMethod    
    paramset_desc: varchar(128)
    param_set_hash: uuid
    unique index (param_set_hash)
    params: longblob  # dictionary of all applicable parameters
    """

    @classmethod
    def insert_new_params(cls, processing_method: str, paramset_idx: int,
                          paramset_desc: str, params: dict):
        param_dict = {'processing_method': processing_method,
                      'paramset_idx': paramset_idx,
                      'paramset_desc': paramset_desc,
                      'params': params,
                      'param_set_hash': dict_to_uuid(params)}
        q_param = cls & {'param_set_hash': param_dict['param_set_hash']}

        if q_param:  # If the specified param-set already exists
            pname = q_param.fetch1('paramset_idx')
            if pname == paramset_idx:  # If the existed set has the same name: job done
                return
            else:  # If not same name: human error, trying to add the same paramset with different name
                raise dj.DataJointError(
                    'The specified param-set already exists - name: {}'.format(pname))
        else:
            cls.insert1(param_dict)


@schema
class CellCompartment(dj.Lookup):
    definition = """  # Cell compartments that can be imaged
    cell_compartment         : char(16)
    """

    contents = zip(['axon', 'soma', 'bouton'])


@schema
class MaskType(dj.Lookup):
    definition = """ # Possible classifications for a segmented mask
    mask_type        : varchar(16)
    """

    contents = zip(['soma', 'axon', 'dendrite', 'neuropil', 'artefact', 'unknown'])


# -------------- Trigger a processing routine --------------

@schema
class ProcessingTask(dj.Manual):
    definition = """  # Manual table for defining a processing task ready to be run
    -> scan.Scan
    -> ProcessingParamSet
    ---
    processing_output_dir: varchar(255)         #  output directory of the processed scan relative to root data directory
    task_mode='load': enum('load', 'trigger')   # 'load': load computed analysis results, 'trigger': trigger computation
    """


@schema
class Processing(dj.Computed):
    definition = """  # Processing Procedure
    -> ProcessingTask
    ---
    processing_time     : datetime  # time of generation of this set of processed, segmented results
    package_version=''  : varchar(16)
    """

    # Run processing only on Scan with ScanInfo inserted
    @property
    def key_source(self):
        return ProcessingTask & scan.ScanInfo

    def make(self, key):
        task_mode = (ProcessingTask & key).fetch1('task_mode')
        method, imaging_dataset = get_loader_result(key, ProcessingTask)

        if task_mode == 'load':
            if method == 'suite2p':
                if (scan.ScanInfo & key).fetch1('nrois') > 0:
                    raise NotImplementedError(f'Suite2p ingestion error - Unable to handle'
                                              f' ScanImage multi-ROI scanning mode yet')
                suite2p_dataset = imaging_dataset
                key = {**key, 'processing_time': suite2p_dataset.creation_time}
            elif method == 'caiman':
                caiman_dataset = imaging_dataset
                key = {**key, 'processing_time': caiman_dataset.creation_time}
            else:
                raise NotImplementedError('Unknown method: {}'.format(method))
        elif task_mode == 'trigger':
            raise NotImplementedError(f'Automatic triggering of {method} analysis'
                                      f' is not yet supported')
        else:
            raise ValueError(f'Unknown task mode: {task_mode}')

        self.insert1(key)


@schema
class Curation(dj.Manual):
    definition = """  #  Different rounds of curation performed on the processing results of the imaging data (no-curation can also be included here)
    -> Processing
    curation_id: int
    ---
    curation_time: datetime             # time of generation of this set of curated results 
    curation_output_dir: varchar(255)   # output directory of the curated results, relative to root data directory
    manual_curation: bool               # has manual curation been performed on this result?
    curation_note='': varchar(2000)  
    """

    def create1_from_processing_task(self, key, is_curated=False, curation_note=''):
        """
        A convenient function to create a new corresponding "Curation" for a particular "ProcessingTask"
        """
        if key not in Processing():
            raise ValueError(f'No corresponding entry in Processing available for: {key};'
                             f' do `Processing.populate(key)`')

        output_dir = (ProcessingTask & key).fetch1('processing_output_dir')
        method, imaging_dataset = get_loader_result(key, ProcessingTask)

        if method == 'suite2p':
            suite2p_dataset = imaging_dataset
            curation_time = suite2p_dataset.creation_time
        elif method == 'caiman':
            caiman_dataset = imaging_dataset
            curation_time = caiman_dataset.creation_time
        else:
            raise NotImplementedError('Unknown method: {}'.format(method))

        # Synthesize curation_id
        curation_id = dj.U().aggr(self & key, n='ifnull(max(curation_id)+1,1)').fetch1('n')
        self.insert1({**key, 'curation_id': curation_id,
                      'curation_time': curation_time, 'curation_output_dir': output_dir,
                      'manual_curation': is_curated,
                      'curation_note': curation_note})


# -------------- Motion Correction --------------

@schema
class MotionCorrection(dj.Imported):
    definition = """  #  Results of motion correction performed on the imaging data
    -> Curation
    ---
    -> scan.Channel.proj(motion_correct_channel='channel') # channel used for motion correction in this processing task
    """

    class RigidMotionCorrection(dj.Part):
        definition = """  # Details of rigid motion correction performed on the imaging data
        -> master
        ---
        outlier_frames=null : longblob  # mask with true for frames with outlier shifts (already corrected)
        y_shifts            : longblob  # (pixels) y motion correction shifts
        x_shifts            : longblob  # (pixels) x motion correction shifts
        z_shifts=null       : longblob  # (pixels) z motion correction shifts (z-drift) 
        y_std               : float     # (pixels) standard deviation of y shifts across all frames
        x_std               : float     # (pixels) standard deviation of x shifts across all frames
        z_std=null          : float     # (pixels) standard deviation of z shifts across all frames
        """

    class NonRigidMotionCorrection(dj.Part):
        """
        Piece-wise rigid motion correction
        - tile the FOV into multiple 3D blocks/patches
        """
        definition = """  # Details of non-rigid motion correction performed on the imaging data
        -> master
        ---
        outlier_frames=null             : longblob      # mask with true for frames with outlier shifts (already corrected)
        block_height                    : int           # (pixels)
        block_width                     : int           # (pixels)
        block_depth                     : int           # (pixels)
        block_count_y                   : int           # number of blocks tiled in the y direction
        block_count_x                   : int           # number of blocks tiled in the x direction
        block_count_z                   : int           # number of blocks tiled in the z direction
        """

    class Block(dj.Part):
        definition = """  # FOV-tiled blocks used for non-rigid motion correction
        -> master.NonRigidMotionCorrection
        block_id        : int
        ---
        block_y         : longblob  # (y_start, y_end) in pixel of this block
        block_x         : longblob  # (x_start, x_end) in pixel of this block
        block_z         : longblob  # (z_start, z_end) in pixel of this block
        y_shifts        : longblob  # (pixels) y motion correction shifts for every frame
        x_shifts        : longblob  # (pixels) x motion correction shifts for every frame
        z_shifts=null   : longblob  # (pixels) x motion correction shifts for every frame
        y_std           : float     # (pixels) standard deviation of y shifts across all frames
        x_std           : float     # (pixels) standard deviation of x shifts across all frames
        z_std=null      : float     # (pixels) standard deviation of z shifts across all frames
        """

    class Summary(dj.Part):
        definition = """ # Summary images for each field and channel after corrections
        -> master
        -> scan.ScanInfo.Field
        ---
        ref_image               : longblob  # image used as alignment template
        average_image           : longblob  # mean of registered frames
        correlation_image=null  : longblob  # correlation map (computed during cell detection)
        max_proj_image=null     : longblob  # max of registered frames
        """

    def make(self, key):
        method, imaging_dataset = get_loader_result(key, Curation)

        field_keys, _ = (scan.ScanInfo.Field & key).fetch('KEY', 'field_z', order_by='field_z')

        if method == 'suite2p':
            suite2p_dataset = imaging_dataset

            motion_correct_channel = suite2p_dataset.planes[0].alignment_channel

            # ---- iterate through all s2p plane outputs ----
            rigid_correction, nonrigid_correction, nonrigid_blocks = {}, {}, {}
            summary_images = []
            for idx, (plane, s2p) in enumerate(suite2p_dataset.planes.items()):
                # -- rigid motion correction --
                if idx == 0:
                    rigid_correction = {
                        **key,
                        'y_shifts': s2p.ops['yoff'],
                        'x_shifts': s2p.ops['xoff'],
                        'z_shifts': np.full_like(s2p.ops['xoff'], 0),
                        'y_std': np.nanstd(s2p.ops['yoff']),
                        'x_std': np.nanstd(s2p.ops['xoff']),
                        'z_std': np.nan,
                        'outlier_frames': s2p.ops['badframes']}
                else:
                    rigid_correction['y_shifts'] = np.vstack(
                        [rigid_correction['y_shifts'], s2p.ops['yoff']])
                    rigid_correction['y_std'] = np.nanstd(
                        rigid_correction['y_shifts'].flatten())
                    rigid_correction['x_shifts'] = np.vstack(
                        [rigid_correction['x_shifts'], s2p.ops['xoff']])
                    rigid_correction['x_std'] = np.nanstd(
                        rigid_correction['x_shifts'].flatten())
                    rigid_correction['outlier_frames'] = np.logical_or(
                        rigid_correction['outlier_frames'], s2p.ops['badframes'])
                # -- non-rigid motion correction --
                if s2p.ops['nonrigid']:
                    if idx == 0:
                        nonrigid_correction = {
                            **key,
                            'block_height': s2p.ops['block_size'][0],
                            'block_width': s2p.ops['block_size'][1],
                            'block_depth': 1,
                            'block_count_y': s2p.ops['nblocks'][0],
                            'block_count_x': s2p.ops['nblocks'][1],
                            'block_count_z': len(suite2p_dataset.planes),
                            'outlier_frames': s2p.ops['badframes']}
                    else:
                        nonrigid_correction['outlier_frames'] = np.logical_or(
                            nonrigid_correction['outlier_frames'], s2p.ops['badframes'])
                    for b_id, (b_y, b_x, bshift_y, bshift_x) in enumerate(
                            zip(s2p.ops['xblock'], s2p.ops['yblock'],
                                s2p.ops['yoff1'].T, s2p.ops['xoff1'].T)):
                        if b_id in nonrigid_blocks:
                            nonrigid_blocks[b_id]['y_shifts'] = np.vstack(
                                [nonrigid_blocks[b_id]['y_shifts'], bshift_y])
                            nonrigid_blocks[b_id]['y_std'] = np.nanstd(
                                nonrigid_blocks[b_id]['y_shifts'].flatten())
                            nonrigid_blocks[b_id]['x_shifts'] = np.vstack(
                                [nonrigid_blocks[b_id]['x_shifts'], bshift_x])
                            nonrigid_blocks[b_id]['x_std'] = np.nanstd(
                                nonrigid_blocks[b_id]['x_shifts'].flatten())
                        else:
                            nonrigid_blocks[b_id] = {
                                **key, 'block_id': b_id,
                                'block_y': b_y, 'block_x': b_x,
                                'block_z': np.full_like(b_x, plane),
                                'y_shifts': bshift_y, 'x_shifts': bshift_x,
                                'z_shifts': np.full((len(suite2p_dataset.planes),
                                                     len(bshift_x)), 0),
                                'y_std': np.nanstd(bshift_y), 'x_std': np.nanstd(bshift_x),
                                'z_std': np.nan}

                # -- summary images --
                motion_correction_key = (scan.ScanInfo.Field * Curation
                                         & key & field_keys[plane]).fetch1('KEY')
                summary_images.append({**motion_correction_key,
                                       'ref_image': s2p.ref_image,
                                       'average_image': s2p.mean_image,
                                       'correlation_image': s2p.correlation_map,
                                       'max_proj_image': s2p.max_proj_image})

            self.insert1({**key, 'motion_correct_channel': motion_correct_channel})
            if rigid_correction:
                self.RigidMotionCorrection.insert1(rigid_correction)
            if nonrigid_correction:
                self.NonRigidMotionCorrection.insert1(nonrigid_correction)
                self.Block.insert(nonrigid_blocks.values())
            self.Summary.insert(summary_images)
        elif method == 'caiman':
            caiman_dataset = imaging_dataset

            self.insert1({**key, 'motion_correct_channel': caiman_dataset.alignment_channel})

            is3D = caiman_dataset.params.motion['is3D']
            if not caiman_dataset.params.motion['pw_rigid']:
                # -- rigid motion correction --
                rigid_correction = {
                    **key,
                    'x_shifts': caiman_dataset.motion_correction['shifts_rig'][:, 0],
                    'y_shifts': caiman_dataset.motion_correction['shifts_rig'][:, 1],
                    'z_shifts': (caiman_dataset.motion_correction['shifts_rig'][:, 2]
                                 if is3D
                                 else np.full_like(
                        caiman_dataset.motion_correction['shifts_rig'][:, 0], 0)),
                    'x_std': np.nanstd(caiman_dataset.motion_correction['shifts_rig'][:, 0]),
                    'y_std': np.nanstd(caiman_dataset.motion_correction['shifts_rig'][:, 1]),
                    'z_std': (np.nanstd(caiman_dataset.motion_correction['shifts_rig'][:, 2])
                              if is3D
                              else np.nan),
                    'outlier_frames': None}

                self.RigidMotionCorrection.insert1(rigid_correction)
            else:
                # -- non-rigid motion correction --
                nonrigid_correction = {
                    **key,
                    'block_height': (caiman_dataset.params.motion['strides'][0]
                                     + caiman_dataset.params.motion['overlaps'][0]),
                    'block_width': (caiman_dataset.params.motion['strides'][1]
                                    + caiman_dataset.params.motion['overlaps'][1]),
                    'block_depth': (caiman_dataset.params.motion['strides'][2]
                                    + caiman_dataset.params.motion['overlaps'][2]
                                    if is3D else 1),
                    'block_count_x': len(
                        set(caiman_dataset.motion_correction['coord_shifts_els'][:, 0])),
                    'block_count_y': len(
                        set(caiman_dataset.motion_correction['coord_shifts_els'][:, 2])),
                    'block_count_z': (len(
                        set(caiman_dataset.motion_correction['coord_shifts_els'][:, 4]))
                                      if is3D else 1),
                    'outlier_frames': None}

                nonrigid_blocks = []
                for b_id in range(len(caiman_dataset.motion_correction['x_shifts_els'][0, :])):
                    nonrigid_blocks.append(
                        {**key, 'block_id': b_id,
                         'block_x': np.arange(*caiman_dataset.motion_correction[
                                                   'coord_shifts_els'][b_id, 0:2]),
                         'block_y': np.arange(*caiman_dataset.motion_correction[
                                                   'coord_shifts_els'][b_id, 2:4]),
                         'block_z': (np.arange(*caiman_dataset.motion_correction[
                                                    'coord_shifts_els'][b_id, 4:6])
                                     if is3D
                                     else np.full_like(
                             np.arange(*caiman_dataset.motion_correction[
                                            'coord_shifts_els'][b_id, 0:2]), 0)),
                         'x_shifts': caiman_dataset.motion_correction[
                                         'x_shifts_els'][:, b_id],
                         'y_shifts': caiman_dataset.motion_correction[
                                         'y_shifts_els'][:, b_id],
                         'z_shifts': (caiman_dataset.motion_correction[
                                          'z_shifts_els'][:, b_id]
                                      if is3D
                                      else np.full_like(
                             caiman_dataset.motion_correction['x_shifts_els'][:, b_id], 0)),
                         'x_std': np.nanstd(caiman_dataset.motion_correction[
                                                'x_shifts_els'][:, b_id]),
                         'y_std': np.nanstd(caiman_dataset.motion_correction[
                                                'y_shifts_els'][:, b_id]),
                         'z_std': (np.nanstd(caiman_dataset.motion_correction[
                                                 'z_shifts_els'][:, b_id])
                                   if is3D
                                   else np.nan)})

                self.NonRigidMotionCorrection.insert1(nonrigid_correction)
                self.Block.insert(nonrigid_blocks)

            # -- summary images --
            summary_images = [
                {**key, **fkey, 'ref_image': ref_image,
                 'average_image': ave_img,
                 'correlation_image': corr_img,
                 'max_proj_image': max_img}
                for fkey, ref_image, ave_img, corr_img, max_img in zip(
                    field_keys,
                    caiman_dataset.motion_correction['reference_image'].transpose(2, 0, 1)
                    if is3D else caiman_dataset.motion_correction[
                        'reference_image'][...][np.newaxis, ...],
                    caiman_dataset.motion_correction['average_image'].transpose(2, 0, 1)
                    if is3D else caiman_dataset.motion_correction[
                        'average_image'][...][np.newaxis, ...],
                    caiman_dataset.motion_correction['correlation_image'].transpose(2, 0, 1)
                    if is3D else caiman_dataset.motion_correction[
                        'correlation_image'][...][np.newaxis, ...],
                    caiman_dataset.motion_correction['max_image'].transpose(2, 0, 1)
                    if is3D else caiman_dataset.motion_correction[
                        'max_image'][...][np.newaxis, ...])]
            self.Summary.insert(summary_images)
        else:
            raise NotImplementedError('Unknown/unimplemented method: {}'.format(method))

# -------------- Segmentation --------------


@schema
class Segmentation(dj.Computed):
    definition = """ # Different mask segmentations.
    -> Curation
    """

    class Mask(dj.Part):
        definition = """ # A mask produced by segmentation.
        -> master
        mask            : smallint
        ---
        -> scan.Channel.proj(segmentation_channel='channel')  # channel used for segmentation
        mask_npix       : int       # number of pixels in ROIs
        mask_center_x   : int       # center x coordinate in pixel
        mask_center_y   : int       # center y coordinate in pixel
        mask_center_z   : int       # center z coordinate in pixel
        mask_xpix       : longblob  # x coordinates in pixels
        mask_ypix       : longblob  # y coordinates in pixels      
        mask_zpix       : longblob  # z coordinates in pixels        
        mask_weights    : longblob  # weights of the mask at the indices above
        """

    def make(self, key):
        method, imaging_dataset = get_loader_result(key, Curation)

        if method == 'suite2p':
            suite2p_dataset = imaging_dataset

            # ---- iterate through all s2p plane outputs ----
            masks, cells = [], []
            for plane, s2p in suite2p_dataset.planes.items():
                mask_count = len(masks)  # increment mask id from all "plane"
                for mask_idx, (is_cell, cell_prob, mask_stat) in enumerate(zip(
                        s2p.iscell, s2p.cell_prob, s2p.stat)):
                    masks.append({
                        **key, 'mask': mask_idx + mask_count,
                        'segmentation_channel': s2p.segmentation_channel,
                        'mask_npix': mask_stat['npix'],
                        'mask_center_x':  mask_stat['med'][1],
                        'mask_center_y':  mask_stat['med'][0],
                        'mask_center_z': mask_stat.get('iplane', plane),
                        'mask_xpix':  mask_stat['xpix'],
                        'mask_ypix':  mask_stat['ypix'],
                        'mask_zpix': np.full(mask_stat['npix'],
                                             mask_stat.get('iplane', plane)),
                        'mask_weights':  mask_stat['lam']})
                    if is_cell:
                        cells.append({
                            **key,
                            'mask_classification_method': 'suite2p_default_classifier',
                            'mask': mask_idx + mask_count,
                            'mask_type': 'soma', 'confidence': cell_prob})

            self.insert1(key)
            self.Mask.insert(masks, ignore_extra_fields=True)

            if cells:
                MaskClassification.insert1({
                    **key,
                    'mask_classification_method': 'suite2p_default_classifier'},
                    allow_direct_insert=True)
                MaskClassification.MaskType.insert(cells,
                                                   ignore_extra_fields=True,
                                                   allow_direct_insert=True)
        elif method == 'caiman':
            caiman_dataset = imaging_dataset

            # infer "segmentation_channel" - from params if available, else from caiman loader
            params = (ProcessingParamSet * ProcessingTask & key).fetch1('params')
            segmentation_channel = params.get('segmentation_channel',
                                              caiman_dataset.segmentation_channel)

            masks, cells = [], []
            for mask in caiman_dataset.masks:
                masks.append({**key,
                              'segmentation_channel': segmentation_channel,
                              'mask': mask['mask_id'],
                              'mask_npix': mask['mask_npix'],
                              'mask_center_x': mask['mask_center_x'],
                              'mask_center_y': mask['mask_center_y'],
                              'mask_center_z': mask['mask_center_z'],
                              'mask_xpix': mask['mask_xpix'],
                              'mask_ypix': mask['mask_ypix'],
                              'mask_zpix': mask['mask_zpix'],
                              'mask_weights': mask['mask_weights']})
                if caiman_dataset.cnmf.estimates.idx_components is not None:
                    if mask['mask_id'] in caiman_dataset.cnmf.estimates.idx_components:
                        cells.append({
                            **key,
                            'mask_classification_method': 'caiman_default_classifier',
                            'mask': mask['mask_id'], 'mask_type': 'soma'})

            self.insert1(key)
            self.Mask.insert(masks, ignore_extra_fields=True)

            if cells:
                MaskClassification.insert1({
                    **key,
                    'mask_classification_method': 'caiman_default_classifier'},
                    allow_direct_insert=True)
                MaskClassification.MaskType.insert(cells,
                                                   ignore_extra_fields=True,
                                                   allow_direct_insert=True)
        else:
            raise NotImplementedError(f'Unknown/unimplemented method: {method}')


@schema
class MaskClassificationMethod(dj.Lookup):
    definition = """
    mask_classification_method: varchar(48)
    """

    contents = zip(['suite2p_default_classifier',
                    'caiman_default_classifier'])


@schema
class MaskClassification(dj.Computed):
    definition = """
    -> Segmentation
    -> MaskClassificationMethod
    """

    class MaskType(dj.Part):
        definition = """
        -> master
        -> Segmentation.Mask
        ---
        -> MaskType
        confidence=null: float
        """

    def make(self, key):
        pass


# -------------- Activity Trace --------------


@schema
class Fluorescence(dj.Computed):
    definition = """  # fluorescence traces before spike extraction or filtering
    -> Segmentation
    """

    class Trace(dj.Part):
        definition = """
        -> master
        -> Segmentation.Mask
        -> scan.Channel.proj(fluo_channel='channel')  # the channel that this trace comes from         
        ---
        fluorescence                : longblob  # fluorescence trace associated with this mask
        neuropil_fluorescence=null  : longblob  # Neuropil fluorescence trace
        """

    def make(self, key):
        method, imaging_dataset = get_loader_result(key, Curation)

        if method == 'suite2p':
            suite2p_dataset = imaging_dataset

            # ---- iterate through all s2p plane outputs ----
            fluo_traces, fluo_chn2_traces = [], []
            for s2p in suite2p_dataset.planes.values():
                mask_count = len(fluo_traces)  # increment mask id from all "plane"
                for mask_idx, (f, fneu) in enumerate(zip(s2p.F, s2p.Fneu)):
                    fluo_traces.append({
                        **key, 'mask': mask_idx + mask_count,
                        'fluo_channel': 0,
                        'fluorescence': f,
                        'neuropil_fluorescence': fneu})
                if len(s2p.F_chan2):
                    mask_chn2_count = len(fluo_chn2_traces) # increment mask id from all planes
                    for mask_idx, (f2, fneu2) in enumerate(zip(s2p.F_chan2, s2p.Fneu_chan2)):
                        fluo_chn2_traces.append({
                            **key, 'mask': mask_idx + mask_chn2_count,
                            'fluo_channel': 1,
                            'fluorescence': f2,
                            'neuropil_fluorescence': fneu2})

            self.insert1(key)
            self.Trace.insert(fluo_traces + fluo_chn2_traces)
        elif method == 'caiman':
            caiman_dataset = imaging_dataset

            # infer "segmentation_channel" - from params if available, else from caiman loader
            params = (ProcessingParamSet * ProcessingTask & key).fetch1('params')
            segmentation_channel = params.get('segmentation_channel',
                                              caiman_dataset.segmentation_channel)

            fluo_traces = []
            for mask in caiman_dataset.masks:
                fluo_traces.append({**key, 'mask': mask['mask_id'],
                                    'fluo_channel': segmentation_channel,
                                    'fluorescence': mask['inferred_trace']})

            self.insert1(key)
            self.Trace.insert(fluo_traces)
        
        else:
            raise NotImplementedError('Unknown/unimplemented method: {}'.format(method))


@schema
class ActivityExtractionMethod(dj.Lookup):
    definition = """
    extraction_method: varchar(32)
    """

    contents = zip(['suite2p_deconvolution', 'caiman_deconvolution', 'caiman_dff'])


@schema
class Activity(dj.Computed):
    definition = """  # inferred neural activity from fluorescence trace - e.g. dff, spikes
    -> Fluorescence
    -> ActivityExtractionMethod
    """

    class Trace(dj.Part):
        definition = """  #
        -> master
        -> Fluorescence.Trace
        ---
        activity_trace: longblob  # 
        """

    @property
    def key_source(self):
        suite2p_key_source = (Fluorescence * ActivityExtractionMethod
                              * ProcessingParamSet.proj('processing_method')
                              & 'processing_method = "suite2p"'
                              & 'extraction_method LIKE "suite2p%"')
        caiman_key_source = (Fluorescence * ActivityExtractionMethod
                             * ProcessingParamSet.proj('processing_method')
                             & 'processing_method = "caiman"'
                             & 'extraction_method LIKE "caiman%"')
        return suite2p_key_source.proj() + caiman_key_source.proj()

    def make(self, key):
        method, imaging_dataset = get_loader_result(key, Curation)

        if method == 'suite2p':
            if key['extraction_method'] == 'suite2p_deconvolution':
                suite2p_dataset = imaging_dataset
                # ---- iterate through all s2p plane outputs ----
                spikes = []
                for s2p in suite2p_dataset.planes.values():
                    mask_count = len(spikes)  # increment mask id from all "plane"
                    for mask_idx, spks in enumerate(s2p.spks):
                        spikes.append({**key, 'mask': mask_idx + mask_count,
                                       'fluo_channel': 0,
                                       'activity_trace': spks})

                self.insert1(key)
                self.Trace.insert(spikes)
        elif method == 'caiman':
            caiman_dataset = imaging_dataset

            if key['extraction_method'] in ('caiman_deconvolution', 'caiman_dff'):
                attr_mapper = {'caiman_deconvolution': 'spikes', 'caiman_dff': 'dff'}

                # infer "segmentation_channel" - from params if available, else from caiman loader
                params = (ProcessingParamSet * ProcessingTask & key).fetch1('params')
                segmentation_channel = params.get('segmentation_channel',
                                                  caiman_dataset.segmentation_channel)

                activities = []
                for mask in caiman_dataset.masks:
                    activities.append({
                        **key, 'mask': mask['mask_id'],
                        'fluo_channel': segmentation_channel,
                        'activity_trace': mask[attr_mapper[key['extraction_method']]]})
                self.insert1(key)
                self.Trace.insert(activities)
        else:
            raise NotImplementedError('Unknown/unimplemented method: {}'.format(method))

# ---------------- HELPER FUNCTIONS ----------------


_table_attribute_mapper = {'ProcessingTask': 'processing_output_dir',
                           'Curation': 'curation_output_dir'}


def get_loader_result(key, table):
    """
    Retrieve the loaded processed imaging results from the loader (e.g. suite2p, caiman, etc.)
        :param key: the `key` to one entry of ProcessingTask or Curation
        :param table: the class defining the table to retrieve
         the loaded results from (e.g. ProcessingTask, Curation)
        :return: a loader object of the loaded results
         (e.g. suite2p.Suite2p, caiman.CaImAn, etc.)
    """
    method, output_dir = (ProcessingParamSet * table & key).fetch1(
        'processing_method', _table_attribute_mapper[table.__name__])

    output_dir = find_full_path(get_imaging_root_data_dir(), output_dir)

    if method == 'suite2p':
        from .readers import suite2p_loader
        loaded_dataset = suite2p_loader.Suite2p(output_dir)
    elif method == 'caiman':
        from .readers import caiman_loader
        loaded_dataset = caiman_loader.CaImAn(output_dir)
    else:
        raise NotImplementedError('Unknown/unimplemented method: {}'.format(method))

    return method, loaded_dataset


def dict_to_uuid(key):
    """
    Given a dictionary `key`, returns a hash string as UUID
    """
    hashed = hashlib.md5()
    for k, v in sorted(key.items()):
        hashed.update(str(k).encode())
        hashed.update(str(v).encode())
    return uuid.UUID(hex=hashed.hexdigest())



"""
ScanImage scans
"""

import datajoint as dj
import pathlib
import importlib
import inspect
from . import find_root_directory

schema = dj.schema()

_linking_module = None


def activate(scan_schema_name, *, create_schema=True, create_tables=True, linking_module=None):
    """
    activate(scan_schema_name, *, create_schema=True, create_tables=True, linking_module=None)
        :param scan_schema_name: schema name on the database server to activate the `scan` module
        :param create_schema: when True (default), create schema in the database if it does not yet exist.
        :param create_tables: when True (default), create tables in the database if they do not yet exist.
        :param linking_module: a module name or a module containing the
         required dependencies to activate the `scan` module:
            Upstream tables:
                + Session: parent table to Scan, typically identifying a recording session
                + Equipment: Reference table for Scan, specifying the equipment used for the acquisition of this scan
                + Location: Reference table for ScanLocation, specifying the brain location where this scan is acquired
            Functions:
                + get_imaging_root_data_dir() -> list
                    Retrieve the full path for the root data directory - e.g. containing the imaging recording files or analysis results for all subject/sessions.
                    :return: a string (or list of string) for full path to the root data directory
                + get_scan_image_files(scan_key: dict) -> list
                    Retrieve the list of ScanImage files associated with a given Scan
                    :param scan_key: key of a Scan
                    :return: list of ScanImage files' full file-paths
                + get_scan_box_files(scan_key: dict) -> list
                    Retrieve the list of Scanbox files (*.sbx) associated with a given Scan
                    :param scan_key: key of a Scan
                    :return: list of Scanbox files' full file-paths
    """

    if isinstance(linking_module, str):
        linking_module = importlib.import_module(linking_module)
    assert inspect.ismodule(linking_module),\
        "The argument 'dependency' must be a module's name or a module"

    global _linking_module
    _linking_module = linking_module

    # activate
    schema.activate(scan_schema_name, create_schema=create_schema,
                    create_tables=create_tables, add_objects=_linking_module.__dict__)


# -------------- Functions required by the element-calcium-imaging  ---------------


def get_imaging_root_data_dir() -> str:
    """
    All data paths, directories in DataJoint Elements are recommended to be stored as
    relative paths (posix format), with respect to some user-configured "root" directory,
     which varies from machine to machine (e.g. different mounted drive locations)

    get_imaging_root_data_dir() -> list
        This user-provided function retrieves the possible root data directories
         containing the imaging data for all subjects/sessions
         (e.g. acquired ScanImage raw files,
         output files from processing routines, etc.)
        :return: a string for full path to the imaging root data directory,
         or list of strings for possible root data directories
    """
    return _linking_module.get_imaging_root_data_dir()


def get_scan_image_files(scan_key: dict) -> list:
    """
    Retrieve the list of ScanImage files associated with a given Scan
    :param scan_key: key of a Scan
    :return: list of ScanImage files' full file-paths
    """
    return _linking_module.get_scan_image_files(scan_key)


def get_scan_box_files(scan_key: dict) -> list:
    """
    Retrieve the list of Scanbox files (*.sbx) associated with a given Scan
    :param scan_key: key of a Scan
    :return: list of Scanbox files' full file-paths
    """
    return _linking_module.get_scan_box_files(scan_key)


# ----------------------------- Table declarations ----------------------


@schema
class AcquisitionSoftware(dj.Lookup):
    definition = """  # Name of acquisition software - e.g. ScanImage, Scanbox
    acq_software: varchar(24)    
    """
    contents = zip(['ScanImage', 'Scanbox'])


@schema
class Channel(dj.Lookup):
    definition = """  # A recording channel
    channel     : tinyint  # 0-based indexing
    """
    contents = zip(range(5))


# ------------ ScanImage's scan ------------


@schema
class Scan(dj.Manual):
    definition = """    
    -> Session
    scan_id: int        
    ---
    -> Equipment  
    -> AcquisitionSoftware  
    scan_notes='' : varchar(4095)         # free-notes
    """


@schema
class ScanLocation(dj.Manual):
    definition = """
    -> Scan   
    ---    
    -> Location      
    """


@schema
class ScanInfo(dj.Imported):
    definition = """ # general data about the reso/meso scans, from ScanImage header
    -> Scan
    ---
    nfields              : tinyint   # number of fields
    nchannels            : tinyint   # number of channels
    ndepths              : int       # Number of scanning depths (planes)
    nframes              : int       # number of recorded frames
    nrois                : tinyint   # number of ROIs (see scanimage's multi ROI imaging)
    x                    : float     # (um) ScanImage's 0 point in the motor coordinate system
    y                    : float     # (um) ScanImage's 0 point in the motor coordinate system
    z                    : float     # (um) ScanImage's 0 point in the motor coordinate system
    fps                  : float     # (Hz) frames per second - Volumetric Scan Rate 
    bidirectional        : boolean   # true = bidirectional scanning
    usecs_per_line=null  : float     # microseconds per scan line
    fill_fraction=null   : float     # raster scan temporal fill fraction (see scanimage)
    """

    class Field(dj.Part):
        definition = """ # field-specific scan information
        -> master
        field_idx         : int
        ---
        px_height         : smallint  # height in pixels
        px_width          : smallint  # width in pixels
        um_height=null    : float     # height in microns
        um_width=null     : float     # width in microns
        field_x           : float     # (um) center of field in the motor coordinate system
        field_y           : float     # (um) center of field in the motor coordinate system
        field_z           : float     # (um) relative depth of field
        delay_image=null  : longblob  # (ms) delay between the start of the scan and pixels in this field
        roi=null          : int       # the scanning roi (as recorded in the acquisition software) containing this field - only relevant to mesoscale scans
        """

    class ScanFile(dj.Part):
        definition = """
        -> master
        file_path: varchar(255)  # filepath relative to root data directory
        """

    def make(self, key):
        """ Read and store some scan meta information."""
        acq_software = (Scan & key).fetch1('acq_software')

        if acq_software == 'ScanImage':
            import scanreader
            # Read the scan
            scan_filepaths = get_scan_image_files(key)
            scan = scanreader.read_scan(scan_filepaths)

            # Insert in ScanInfo
            x_zero, y_zero, z_zero = scan.motor_position_at_zero  # motor x, y, z at ScanImage's 0
            self.insert1(dict(key,
                              nfields=scan.num_fields,
                              nchannels=scan.num_channels,
                              nframes=scan.num_frames,
                              ndepths=scan.num_scanning_depths,
                              x=x_zero,
                              y=y_zero,
                              z=z_zero,
                              fps=scan.fps,
                              bidirectional=scan.is_bidirectional,
                              usecs_per_line=scan.seconds_per_line * 1e6,
                              fill_fraction=scan.temporal_fill_fraction,
                              nrois=scan.num_rois if scan.is_multiROI else 0))

            # Insert Field(s)
            if scan.is_multiROI:
                self.Field.insert([
                    dict(key,
                         field_idx=field_id,
                         px_height=scan.field_heights[field_id],
                         px_width=scan.field_widths[field_id],
                         um_height=scan.field_heights_in_microns[field_id],
                         um_width=scan.field_widths_in_microns[field_id],
                         field_x=x_zero + scan._degrees_to_microns(scan.fields[field_id].x),
                         field_y=y_zero + scan._degrees_to_microns(scan.fields[field_id].y),
                         field_z=z_zero + scan.fields[field_id].depth,
                         delay_image=scan.field_offsets[field_id],
                         roi=scan.field_rois[field_id][0])
                    for field_id in range(scan.num_fields)])
            else:
                self.Field.insert([
                    dict(key,
                         field_idx=plane_idx,
                         px_height=scan.image_height,
                         px_width=scan.image_width,
                         um_height=getattr(scan, 'image_height_in_microns', None),
                         um_width=getattr(scan, 'image_width_in_microns', None),
                         field_x=x_zero,
                         field_y=y_zero,
                         field_z=z_zero + scan.scanning_depths[plane_idx],
                         delay_image=scan.field_offsets[plane_idx])
                    for plane_idx in range(scan.num_scanning_depths)])
        elif acq_software == 'Scanbox':
            import sbxreader
            # Read the scan
            scan_filepaths = get_scan_box_files(key)
            sbx_meta = sbxreader.sbx_get_metadata(scan_filepaths[0])
            sbx_matinfo = sbxreader.sbx_get_info(scan_filepaths[0])
            is_multiROI = bool(sbx_matinfo.mesoscope.enabled)  # currently not handling "multiROI" ingestion

            if is_multiROI:
                raise NotImplementedError(
                    'Loading routine not implemented for Scanbox multiROI scan mode')

            # Insert in ScanInfo
            x_zero, y_zero, z_zero = sbx_meta['stage_pos'] 
            self.insert1(dict(key,
                              nfields=sbx_meta['num_fields']
                              if is_multiROI else sbx_meta['num_planes'],
                              nchannels=sbx_meta['num_channels'],
                              nframes=sbx_meta['num_frames'],
                              ndepths=sbx_meta['num_planes'],
                              x=x_zero,
                              y=y_zero,
                              z=z_zero,
                              fps=sbx_meta['frame_rate'],
                              bidirectional=sbx_meta == 'bidirectional',
                              nrois=sbx_meta['num_rois'] if is_multiROI else 0))
            # Insert Field(s)
            if not is_multiROI:
                px_width, px_height = sbx_meta['frame_size']
                self.Field.insert([dict(key,
                                        field_idx=plane_idx,
                                        px_height=px_height,
                                        px_width=px_width,
                                        um_height=px_height * sbx_meta['um_per_pixel_y']
                                        if sbx_meta['um_per_pixel_y'] else None,
                                        um_width=px_width * sbx_meta['um_per_pixel_x']
                                        if sbx_meta['um_per_pixel_x'] else None,
                                        field_x=x_zero,
                                        field_y=y_zero,
                                        field_z=z_zero + sbx_meta['etl_pos'][plane_idx])
                                   for plane_idx in range(sbx_meta['num_planes'])])

        else:
            raise NotImplementedError(
                f'Loading routine not implemented for {acq_software} acquisition software')

        # Insert file(s)
        root_dir = find_root_directory(get_imaging_root_data_dir(), scan_filepaths[0])

        scan_files = [pathlib.Path(f).relative_to(root_dir).as_posix() for f in scan_filepaths]
        self.ScanFile.insert([{**key, 'file_path': f} for f in scan_files])



import h5py
import caiman as cm
import scipy
import numpy as np
from datetime import datetime
import os
import pathlib
from tqdm import tqdm


_required_hdf5_fields = ['/motion_correction/reference_image',
                         '/motion_correction/correlation_image',
                         '/motion_correction/average_image',
                         '/motion_correction/max_image',
                         '/estimates/A']


class CaImAn:
    """
    Parse the CaImAn output file
    Expecting the following objects:
    - 'dims':
    - 'dview':
    - 'estimates':              Segmentations and traces
    - 'mmap_file':
    - 'params':                 Input parameters
    - 'remove_very_bad_comps':
    - 'skip_refinement':
    - 'motion_correction':      Motion correction shifts and summary images
    CaImAn results doc: https://caiman.readthedocs.io/en/master/Getting_Started.html#result-variables-for-2p-batch-analysis
    """

    def __init__(self, caiman_dir):
        # ---- Search and verify CaImAn output file exists ----
        caiman_dir = pathlib.Path(caiman_dir)
        if not caiman_dir.exists():
            raise FileNotFoundError('CaImAn directory not found: {}'.format(caiman_dir))

        for fp in caiman_dir.glob('*.hdf5'):
            with h5py.File(fp, 'r') as h5f:
                if all(s in h5f for s in _required_hdf5_fields):
                    self.caiman_fp = fp
                    break
        else:
            raise FileNotFoundError(
                'No CaImAn analysis output file found at {}'
                ' containg all required fields ({})'.format(caiman_dir, _required_hdf5_fields))

        # ---- Initialize CaImAn's results ----
        self.cnmf = cm.source_extraction.cnmf.cnmf.load_CNMF(self.caiman_fp)
        self.params = self.cnmf.params

        self.h5f = h5py.File(self.caiman_fp, 'r')
        self.motion_correction = self.h5f['motion_correction']
        self._masks = None

        # ---- Metainfo ----
        self.creation_time = datetime.fromtimestamp(os.stat(self.caiman_fp).st_ctime)
        self.curation_time = datetime.fromtimestamp(os.stat(self.caiman_fp).st_ctime)

    @property
    def masks(self):
        if self._masks is None:
            self._masks = self.extract_masks()
        return self._masks

    @property
    def alignment_channel(self):
        return 0  # hard-code to channel index 0

    @property
    def segmentation_channel(self):
        return 0  # hard-code to channel index 0

    def extract_masks(self):
        if self.params.motion['is3D']:
            raise NotImplemented('CaImAn mask extraction for volumetric data not yet implemented')

        comp_contours = cm.utils.visualization.get_contours(
            self.cnmf.estimates.A, self.cnmf.dims)

        masks = []
        for comp_idx, comp_contour in enumerate(comp_contours):
            ind, _, weights = scipy.sparse.find(self.cnmf.estimates.A[:, comp_idx])
            if self.cnmf.params.motion['is3D']:
                xpix, ypix, zpix = np.unravel_index(ind, self.cnmf.dims, order='F')
                center_x, center_y, center_z = comp_contour['CoM'].astype(int)
            else:
                xpix, ypix = np.unravel_index(ind, self.cnmf.dims, order='F')
                center_x, center_y = comp_contour['CoM'].astype(int)
                center_z = 0
                zpix = np.full(len(weights), center_z)

            masks.append({'mask_id': comp_contour['neuron_id'],
                          'mask_npix': len(weights), 'mask_weights': weights,
                          'mask_center_x': center_x,
                          'mask_center_y': center_y,
                          'mask_center_z': center_z,
                          'mask_xpix': xpix, 'mask_ypix': ypix, 'mask_zpix': zpix,
                          'inferred_trace': self.cnmf.estimates.C[comp_idx, :],
                          'dff': self.cnmf.estimates.F_dff[comp_idx, :],
                          'spikes': self.cnmf.estimates.S[comp_idx, :]})
        return masks


def _process_scanimage_tiff(scan_filenames, output_dir='./'):
    """
    Read ScanImage TIFF - reshape into volumetric data based on scanning depths and channels
    Save new TIFF files for each channel - with shape (frame x height x width x depth)
    """
    from skimage.external.tifffile import imsave
    import scanreader

    # ------------ CaImAn multi-channel multi-plane tiff file ------------
    for scan_filename in tqdm(scan_filenames):
        scan = scanreader.read_scan(scan_filename)
        cm_movie = cm.load(scan_filename)

        # ---- Volumetric movie: (depth x height x width x channel x frame) ----
        # tiff pages are ordered as:
        # ch0-pln0-t0, ch1-pln0-t0, ch0-pln1-t0, ch1-pln1-t0, ..., ch0-pln1-t5, ch1-pln1-t5, ...

        vol_timeseries = np.full((scan.num_scanning_depths,
                                  scan.image_height, scan.image_width,
                                scan.num_channels, scan.num_frames), 0).astype(scan.dtype)
        for pln_idx in range(scan.num_scanning_depths):
            for chn_idx in range(scan.num_channels):
                pln_chn_ind = np.arange(pln_idx * scan.num_channels + chn_idx,
                                        scan._num_pages,
                                        scan.num_scanning_depths * scan.num_channels)
                vol_timeseries[pln_idx, :, :, chn_idx, :] = cm_movie[pln_chn_ind, :, :].transpose(1, 2, 0)

        # save volumetric movie for individual channel
        output_dir = pathlib.Path(output_dir)
        fname = pathlib.Path(scan_filename).stem

        for chn_idx in range(scan.num_channels):
            if scan.num_scanning_depths == 1:
                chn_vol = vol_timeseries[0, :, :, chn_idx, :].squeeze().transpose(2, 0, 1)  # (frame x height x width)
            else:
                chn_vol = vol_timeseries[:, :, :, chn_idx, :].transpose(3, 1, 2, 0)  # (frame x height x width x depth)
            save_fp = output_dir / '{}_chn{}.tif'.format(fname, chn_idx)
            imsave(save_fp.as_posix(), chn_vol)


def _save_mc(mc, caiman_fp, is3D):
    """
    DataJoint Imaging Element - CaImAn Integration
    Run these commands after the CaImAn analysis has completed.
    This will save the relevant motion correction data into the '*.hdf5' file.
    Please do not clear variables from memory prior to running these commands.
    The motion correction (mc) object will be read from memory.

    'mc' :                CaImAn motion correction object
    'caiman_fp' :         CaImAn output (*.hdf5) file path

    'shifts_rig' :        Rigid transformation x and y shifts per frame
    'x_shifts_els' :      Non rigid transformation x shifts per frame per block
    'y_shifts_els' :      Non rigid transformation y shifts per frame per block
    """

    # Load motion corrected mmap image
    mc_image = cm.load(mc.mmap_file, is3D=is3D)

    # Compute motion corrected summary images
    average_image = np.mean(mc_image, axis=0)
    max_image = np.max(mc_image, axis=0)

    # Compute motion corrected correlation image
    correlation_image = cm.local_correlations(mc_image.transpose((1, 2, 3, 0)
                                                                 if is3D else (1, 2, 0)))
    correlation_image[np.isnan(correlation_image)] = 0

    # Compute mc.coord_shifts_els
    grid = []
    if is3D:
        for _, _, _, x, y, z, _ in cm.motion_correction.sliding_window_3d(
                mc_image[0, :, :, :], mc.overlaps, mc.strides):
            grid.append([x, x + mc.overlaps[0] + mc.strides[0],
                         y, y + mc.overlaps[1] + mc.strides[1],
                         z, z + mc.overlaps[2] + mc.strides[2]])
    else:
        for _, _, x, y, _ in cm.motion_correction.sliding_window(
                mc_image[0, :, :], mc.overlaps, mc.strides):
            grid.append([x, x + mc.overlaps[0] + mc.strides[0],
                         y, y + mc.overlaps[1] + mc.strides[1]])

    # Open hdf5 file and create 'motion_correction' group
    h5f = h5py.File(caiman_fp, 'r+')
    h5g = h5f.require_group("motion_correction")

    # Write motion correction shifts and motion corrected summary images to hdf5 file
    if mc.pw_rigid:
        h5g.require_dataset("x_shifts_els", shape=np.shape(mc.x_shifts_els),
                            data=mc.x_shifts_els,
                            dtype=mc.x_shifts_els[0][0].dtype)
        h5g.require_dataset("y_shifts_els", shape=np.shape(mc.y_shifts_els),
                            data=mc.y_shifts_els,
                            dtype=mc.y_shifts_els[0][0].dtype)
        if is3D:
            h5g.require_dataset("z_shifts_els", shape=np.shape(mc.z_shifts_els),
                                data=mc.z_shifts_els,
                                dtype=mc.z_shifts_els[0][0].dtype)

        h5g.require_dataset("coord_shifts_els", shape=np.shape(grid),
                            data=grid, dtype=type(grid[0][0]))

        # For CaImAn, reference image is still a 2D array even for the case of 3D
        # Assume that the same ref image is used for all the planes
        reference_image = np.tile(mc.total_template_els, (1, 1, correlation_image.shape[-1]))\
            if is3D else mc.total_template_els
    else:
        h5g.require_dataset("shifts_rig", shape=np.shape(mc.shifts_rig),
                            data=mc.shifts_rig, dtype=mc.shifts_rig[0][0].dtype)
        h5g.require_dataset("coord_shifts_rig", shape=np.shape(grid),
                            data=grid, dtype=type(grid[0][0]))
        reference_image = np.tile(mc.total_template_rig, (1, 1, correlation_image.shape[-1]))\
            if is3D else mc.total_template_rig

    h5g.require_dataset("reference_image", shape=np.shape(reference_image),
                        data=reference_image,
                        dtype=reference_image.dtype)
    h5g.require_dataset("correlation_image", shape=np.shape(correlation_image),
                        data=correlation_image,
                        dtype=correlation_image.dtype)
    h5g.require_dataset("average_image", shape=np.shape(average_image),
                        data=average_image, dtype=average_image.dtype)
    h5g.require_dataset("max_image", shape=np.shape(max_image),
                        data=max_image, dtype=max_image.dtype)

    # Close hdf5 file
    h5f.close()



import numpy as np
import pathlib
from datetime import datetime
from collections import OrderedDict


_suite2p_ftypes = ('ops', 'Fneu', 'Fneu_chan2', 'F', 'F_chan2',
                   'iscell', 'spks', 'stat', 'redcell')


class Suite2p:
    """
    Wrapper class containing all suite2p outputs from one suite2p analysis routine.
    This includes outputs from the individual plane, with plane indexing starting from 0
    Plane index of -1 indicates a suite2p "combined" outputs from all planes,
     thus saved in the "planes_combined" attribute

    A 'suite2p_dir' example:

        -- suite2p_dir
          -- plane0
             -- ops.npy
             -- F.npy
             -- ...
          -- plane1
             -- ops.npy
             -- F.npy
             -- ...
          -- combined
             -- ops.npy
             -- F.npy
             -- ...
    """

    def __init__(self, suite2p_dir):
        self.suite2p_dir = pathlib.Path(suite2p_dir)

        ops_filepaths = list(self.suite2p_dir.rglob('*ops.npy'))

        if not len(ops_filepaths):
            raise FileNotFoundError(
                'Suite2p output result files not found at {}'.format(suite2p_dir))

        self.planes = {}
        self.planes_combined = None
        for ops_fp in ops_filepaths:
            plane_s2p = PlaneSuite2p(ops_fp.parent)
            if plane_s2p.plane_idx == -1:
                self.planes_combined = plane_s2p
            else:
                self.planes[plane_s2p.plane_idx] = plane_s2p
        self.planes = OrderedDict({k: self.planes[k] for k in sorted(self.planes)})

        self.creation_time = min([p.creation_time for p in self.planes.values()])  # ealiest file creation time
        self.curation_time = max([p.curation_time for p in self.planes.values()])  # most recent curation time


class PlaneSuite2p:
    """
    Parse the suite2p output directory and load data, ***per plane***.
    Expecting the following files:
    - 'ops':        Options file
    - 'Fneu':       Neuropil traces file for functional channel
    - 'Fneu_chan2': Neuropil traces file for channel 2
    - 'F':          Fluorescence traces for functional channel
    - 'F_chan2':    Fluorescence traces for channel 2
    - 'iscell':     Array of (user curated) cells and probability of being a cell
    - 'spks':       Spikes (raw deconvolved with OASIS package)
    - 'stat':       Various statistics for each cell
    - 'redcell':    "Red cell" (second channel) stats

    Suite2p output doc: https://suite2p.readthedocs.io/en/latest/outputs.html
    """

    def __init__(self, suite2p_plane_dir):
        self.fpath = pathlib.Path(suite2p_plane_dir)

        # ---- Verify dataset exists ----
        ops_fp = self.fpath / 'ops.npy'
        if not ops_fp.exists():
            raise FileNotFoundError(
                'No "ops.npy" found. Invalid suite2p plane folder: {}'.format(self.fpath))
        self.creation_time = datetime.fromtimestamp(ops_fp.stat().st_ctime)

        iscell_fp = self.fpath / 'iscell.npy'
        if not iscell_fp.exists():
            raise FileNotFoundError(
                'No "iscell.npy" found. Invalid suite2p plane folder: {}'.format(self.fpath))
        self.curation_time = datetime.fromtimestamp(iscell_fp.stat().st_ctime)

        # ---- Initialize attributes ----
        for s2p_type in _suite2p_ftypes:
            setattr(self, '_{}'.format(s2p_type), None)
        self._cell_prob = None

        self.plane_idx = (-1 if self.fpath.name == 'combined'
                          else int(self.fpath.name.replace('plane', '')))

    # ---- load core files ----

    @property
    def ops(self):
        if self._ops is None:
            fp = self.fpath / 'ops.npy'
            self._ops = np.load(fp, allow_pickle=True).item()
        return self._ops

    @property
    def Fneu(self):
        if self._Fneu is None:
            fp = self.fpath / 'Fneu.npy'
            self._Fneu = np.load(fp) if fp.exists() else []
        return self._Fneu

    @property
    def Fneu_chan2(self):
        if self._Fneu_chan2 is None:
            fp = self.fpath / 'Fneu_chan2.npy'
            self._Fneu_chan2 = np.load(fp) if fp.exists() else []
        return self._Fneu_chan2

    @property
    def F(self):
        if self._F is None:
            fp = self.fpath / 'F.npy'
            self._F = np.load(fp) if fp.exists() else []
        return self._F

    @property
    def F_chan2(self):
        if self._F_chan2 is None:
            fp = self.fpath / 'F_chan2.npy'
            self._F_chan2 = np.load(fp) if fp.exists() else []
        return self._F_chan2

    @property
    def iscell(self):
        if self._iscell is None:
            fp = self.fpath / 'iscell.npy'
            d = np.load(fp)
            self._iscell = d[:, 0].astype(bool)
            self._cell_prob = d[:, 1]
        return self._iscell

    @property
    def cell_prob(self):
        if self._cell_prob is None:
            fp = self.fpath / 'iscell.npy'
            if fp.exists():
                d = np.load(fp)
                self._iscell = d[:, 0].astype(bool)
                self._cell_prob = d[:, 1]
        return self._cell_prob

    @property
    def spks(self):
        if self._spks is None:
            fp = self.fpath / 'spks.npy'
            self._spks = np.load(fp) if fp.exists() else []
        return self._spks

    @property
    def stat(self):
        if self._stat is None:
            fp = self.fpath / 'stat.npy'
            self._stat = np.load(fp, allow_pickle=True) if fp.exists() else []
        return self._stat

    @property
    def redcell(self):
        if self._redcell is None:
            fp = self.fpath / 'redcell.npy'
            self._redcell = np.load(fp) if fp.exists() else []
        return self._redcell

    # ---- image property ----

    @property
    def ref_image(self):
        return self.ops['refImg']

    @property
    def mean_image(self):
        return self.ops['meanImg']

    @property
    def max_proj_image(self):
        return self.ops.get('max_proj', np.full_like(self.mean_image, np.nan))

    @property
    def correlation_map(self):
        return self.ops['Vcorr']

    @property
    def alignment_channel(self):
        return self.ops['align_by_chan'] - 1  # suite2p is 1-based, convert to 0-based

    @property
    def segmentation_channel(self):
        return self.ops['functional_chan'] - 1  # suite2p is 1-based, convert to 0-based



import re
from datetime import datetime


def parse_scanimage_header(scan):
    header = {}
    for item in scan.header.split('\n'):
        try:
            key, value = item.split(' = ')
            key = re.sub('^scanimage_', '', key.replace('.', '_'))
            header[key] = value
        except:
            pass
    return header


def get_scanimage_acq_time(scan):
    header = parse_scanimage_header(scan)
    recording_time = datetime.strptime((header['epoch'][1:-1]).replace(',', ' '),
                                       '%Y %m %d %H %M %S.%f')
    return recording_time
